from myst_libre.tools import JupyterHubLocalSpawner, MystMD
from myst_libre.abstract_class import AbstractClass

class MystBuilder(AbstractClass):
    def __init__(self, hub=None, build_dir=None):
        if hub is not None:
            if not isinstance(hub, JupyterHubLocalSpawner):
                raise TypeError(f"Expected 'hub' to be an instance of JupyterHubLocalSpawner, got {type(hub).__name__} instead")
            self.hub = hub
            self.env_vars = {
                "JUPYTER_BASE_URL": f"{self.hub.jh_url}",
                "JUPYTER_TOKEN": f"{self.hub.jh_token}",
                "port": f"{self.hub.port}"
            }
            self.build_dir = self.hub.rees.build_dir
        else:
            if build_dir is None:
                raise ValueError("If 'hub' is None, 'build_dir' must be provided")
            self.build_dir = build_dir
            self.env_vars = {}
            self.hub = None

        super().__init__()
        self.myst_client = MystMD(self.build_dir, self.env_vars)
    
    def setenv(self,key,value):
        self.env_vars[key] = value

    def build(self,*args,user=None,group=None):
        if self.hub is not None:
            self.cprint(f'Starting MyST build {self.hub.jh_url}','yellow')
        else:
            self.cprint(f'Starting MyST build no exec.','yellow')
        self.myst_client.build('build',*args,user=user,group=group)


from myst_libre.abstract_class import AbstractClass
from myst_libre.tools.docker_registry_client import DockerRegistryClient
from myst_libre.tools.build_source_manager import BuildSourceManager
import docker
import subprocess
import os

class REES(DockerRegistryClient,BuildSourceManager):
    def __init__(self, rees_dict):
        # These are needed in the scope of the base classes
        self.registry_url = rees_dict['registry_url']
        self.gh_user_repo_name = rees_dict['gh_user_repo_name']
        self.gh_repo_commit_hash = rees_dict['gh_repo_commit_hash']
        self.binder_image_tag = rees_dict['binder_image_tag']
        self.binder_image_name = rees_dict.get('binder_image_name', None)
        
        if 'dotenv' in rees_dict.keys():
            self.dotenvloc = rees_dict['dotenv']

        # Initialize as base to rees
        BuildSourceManager.__init__(self)
        DockerRegistryClient.__init__(self)

        self.cprint(f"␤[Preflight checks]","light_grey")
        self.check_docker_installed()

        self.pull_image_name = ""
        self.use_public_registry = False
        self.repo_commit_info = {}
        self.binder_commit_info = {}
        # CHECK: This may not work properly without
        # logging in to the registry on the host machine 
        # which keeps that auth info on the config file. 
        self.docker_client = docker.from_env()
    
    def check_docker_installed(self):
        """
        Check if Docker is installed and available in the system PATH.

        Raises:
            EnvironmentError: If Docker is not installed or not found in PATH.
        """
        try:
            result = subprocess.run(['docker', '--version'], env=os.environ, capture_output=True, text=True, check=True)
            self.cprint(f"✓ Docker is installed: {result.stdout.strip()}",'green')
        except subprocess.CalledProcessError as e:
            raise EnvironmentError("Docker is not installed or not found in PATH. Please install Docker to proceed.") from e

    def login_to_registry(self):
        """
        Login to a private docker registry.
        """
        self.docker_client.login(username=self._auth['username'], password=self._auth['password'], registry=self.registry_url)

    def pull_image(self):
        """
        Pull the Docker image from the registry.
        """
        if bool(self._auth) or not self.use_public_registry:
            self.login_to_registry()
            self.logger.info(f"Logging into {self.registry_url_bare}")

        try:
            self.pull_image_name = f'{self.registry_url_bare}/{self.found_image_name}'
            self.logger.info(f'Pulling image {self.pull_image_name}:{self.binder_image_tag} from {self.registry_url}.')
            self.docker_image = self.docker_client.images.pull(self.pull_image_name, tag=self.binder_image_tag)
        except:
            self.logger.info(f'Pulling image {self.found_image_name}:{self.binder_image_tag} from {self.registry_url}.')
            self.docker_image = self.docker_client.images.pull(self.found_image_name, tag=self.binder_image_tag)


"""
build_source_manager.py

This module contains the BuildSourceManager class for handling source code repositories.
"""

import os
import json
os.environ["GIT_PYTHON_REFRESH"] = "quiet"
from git import Repo
from datetime import datetime
from myst_libre.abstract_class import AbstractClass
from repo2data.repo2data import Repo2Data

class BuildSourceManager(AbstractClass):
    """
    Manager for handling source code repositories.
    
    Args:
        gh_user_repo_name (str): GitHub user/repository name.
        gh_repo_commit_hash (str): Commit hash of the repository.
    """
    def __init__(self):
        super().__init__()
        self.build_dir = ""
        self.branch = 'main'
        self.provider = 'https://github.com'
        self.username = self.gh_user_repo_name.split('/')[0]
        self.repo_name = self.gh_user_repo_name.split('/')[1]
        now = datetime.now()
        self.created_at = now.strftime("%Y-%m-%dT%H:%M:%S")
        self.dataset_name = ""

    def create_build_dir_host(self):
        """
        Create build directory on the host machine.
        
        Returns:
            bool: True if directory created, else False.
        """
        if not os.path.exists(self.build_dir):
            os.makedirs(self.build_dir, exist_ok=True)
            return True
        return False

    def git_clone_repo(self,clone_parent_directory):
        """
        Clone the GitHub repository.
        
        Returns:
            bool: True if cloned successfully, else False.
        """
        self.host_build_source_parent_dir = clone_parent_directory
        self.build_dir = os.path.join(self.host_build_source_parent_dir, self.username, self.repo_name, self.gh_repo_commit_hash)
        
        if os.path.exists(self.build_dir):
            self.cprint(f'Source {self.build_dir} already exists.', "yellow")
            self.repo_object = Repo(self.build_dir)
        else:
            os.makedirs(os.path.dirname(self.build_dir), exist_ok=True)
            self.cprint(f'Cloning into {self.build_dir}', "green")
            self.repo_object = Repo.clone_from(f'{self.provider}/{self.gh_user_repo_name}', self.build_dir)
        
        self.set_commit_info()
        self.validate_commits()

    def git_checkout_commit(self):
        """
        Checkout the specified commit in the repository.
        
        Returns:
            bool: True if checked out successfully.
        """
        self.cprint(f'Checking out {self.gh_repo_commit_hash}', "green")
        self.repo_object.git.checkout(self.gh_repo_commit_hash)
        return True

    def get_project_name(self):
        """
        Get the project name from the data requirement file.
        
        Returns:
            str: Project name.
        """
        data_config_dir = os.path.join(self.build_dir, 'binder', 'data_requirement.json')
        with open(data_config_dir, 'r') as file:
            data = json.load(file)
        self.dataset_name = data['projectName']
    
    def repo2data_download(self,target_directory):
        data_req_path = os.path.join(self.build_dir, 'binder', 'data_requirement.json')
        if not os.path.isfile(data_req_path):
            self.cprint(f'Skipping repo2data download', "yellow")
        else:
            self.cprint(f'Starting repo2data download', "green")
            repo2data = Repo2Data(data_req_path, server=True)
            repo2data.set_server_dst_folder(target_directory)
            repo2data.install()

    def set_commit_info(self):
        if self.binder_image_name:
            self.binder_commit_info['datetime'] = "20 November 2024"
            self.binder_commit_info['message'] =  "Base runtime from myst-libre"
        else:
            self.binder_commit_info['datetime'] = self.repo_object.commit(self.binder_image_tag).committed_datetime
            self.binder_commit_info['message'] = self.repo_object.commit(self.binder_image_tag).message
        self.repo_commit_info['datetime'] = self.repo_object.commit(self.gh_repo_commit_hash).committed_datetime
        self.repo_commit_info['message'] = self.repo_object.commit(self.gh_repo_commit_hash).message
        self.validate_commits()

    def validate_commits(self):
        if not self.binder_image_name:
            if self.repo_commit_info['datetime'] < self.binder_commit_info['datetime']:
                raise ValueError("The repo commit datetime cannot be older than the binder commit datetime.")

    def create_latest_symlink(self):
        """
        Create a symlink to the latest build directory.
        """
        self.latest_dir = os.path.join(self.host_build_source_parent_dir, self.username, self.repo_name, 'latest')
        self.logger.info(f'Creating symlink {self.gh_repo_commit_hash} --> latest')
        if not os.path.exists(self.latest_dir):
            os.makedirs(self.latest_dir)
        else:
            for item in os.listdir(self.build_dir):
                os.unlink(item)
        for item in os.listdir(self.build_dir):
            source_path = os.path.join(self.build_dir, item)
            target_path = os.path.join(self.latest_dir, item)
            if os.path.isdir(source_path):
                os.symlink(source_path, target_path, target_is_directory=True)
            else:
                os.symlink(source_path, target_path)



"""
jupyter_hub_local_spawner.py

This module contains the JupyterHubLocalSpawner class for managing JupyterHub instances locally.
"""

import os
import logging
import socket
from hashlib import blake2b
from myst_libre.abstract_class import AbstractClass
from myst_libre.rees import REES

class JupyterHubLocalSpawner(AbstractClass):
    """
    Spawner for managing JupyterHub instances locally.
    
    Args:
        registry_url (str): URL of the Docker registry (https://my-registry.example.com).
        gh_user_repo_name (str): GitHub user/repository name.
        auth (dict): Authentication credentials {username:"***","password":"***"}.
        binder_image_tag (str): Docker image tag of the container in which the article will be built.
        build_src_commit_hash (str): Commit hash of the repository from which the article will be built.
    """
    def __init__(self,rees,**kwargs):
        if not isinstance(rees, REES):
            raise TypeError(f"Expected 'rees' to be an instance of REES, got {type(rees).__name__} instead")
        super().__init__()
        self.rees = rees
        required_inputs = ['container_data_mount_dir','container_build_source_mount_dir',
                           'host_data_parent_dir','host_build_source_parent_dir']

        for inp in required_inputs:
            if inp not in kwargs.keys():
                raise(f'{inp} is not provided for JupyterHubLocalSpawner.')
            else:
                setattr(self, inp, kwargs[inp])

        self.container = None
        self.port = None
        self.jh_token = None

    def find_open_port(self):
        """
        Find an open port to use.
        
        Returns:
            int: Available port number.
        
        Raises:
            Exception: If no open ports are available.
        """
        for port in range(8888, 10000):
            if not self._is_port_in_use(port):
                return port
        raise Exception("No open ports available")

    def _is_port_in_use(self, port):
        """
        Check if a port is in use.
        
        Args:
            port (int): Port number to check.
        
        Returns:
            bool: True if port is in use, else False.
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(('127.0.0.1', port)) == 0

    def spawn_jupyter_hub(self,jb_build_command=None):
        """
        Spawn a JupyterHub instance.
        """
        self.port = self.find_open_port()
        h = blake2b(digest_size=20)
        h.update(os.urandom(20))
        self.jh_token = h.hexdigest()

        if jb_build_command:
            this_entrypoint = f"/bin/sh -c 'jupyter-book build --all --verbose --path-output {self.container_build_source_mount_dir} content 2>&1 | tee -a jupyter_book_build.log'"
        else:
            this_entrypoint = f'jupyter server --allow-root --ip 0.0.0.0 --log-level=DEBUG --IdentityProvider.token="{self.jh_token}" --ServerApp.port="{self.port}"'

        if not self.rees.search_img_by_repo_name():
            raise Exception(f"[ERROR] A docker image has not been found for {self.rees.gh_user_repo_name} at {self.rees.binder_image_tag}.")
        if self.rees.binder_image_tag not in self.rees.found_image_tags:
            raise Exception(f"[ERROR] A docker image exists for {self.rees.gh_user_repo_name}, yet the tag {self.rees.binder_image_tag} is missing.")
        
        # self.rees.found_image_name is assigned if above not fails

        self.rees.git_clone_repo(self.host_build_source_parent_dir)
        self.rees.git_checkout_commit()
        if not self.rees.dataset_name:
            self.rees.get_project_name()
        
        if self.rees.dataset_name:
            self.rees.repo2data_download(self.host_data_parent_dir)
            mnt_vol = {f'{os.path.join(self.host_data_parent_dir,self.rees.dataset_name)}': {'bind': os.path.join(self.container_data_mount_dir,self.rees.dataset_name), 'mode': 'ro'},
                    self.rees.build_dir: {'bind': f'{self.container_build_source_mount_dir}', 'mode': 'rw'}}
        else:
            mnt_vol = {self.rees.build_dir: {'bind': f'{self.container_build_source_mount_dir}', 'mode': 'rw'}}
            
        self.rees.pull_image()
        self.jh_url = f"http://localhost:{self.port}"
        try:
            self.container = self.rees.docker_client.containers.run(
                self.rees.docker_image,
                ports={f'{self.port}/tcp': self.port},
                environment={"JUPYTER_TOKEN": f'{self.jh_token}',"port": f'{self.port}',"JUPYTER_BASE_URL": f'{self.jh_url}'},
                entrypoint= this_entrypoint,
                volumes=mnt_vol,
                detach=True)
            logging.info(f'Jupyter hub is {self.container.status}')
            self.cprint(f'␤[Status]', 'light_grey')
            self.cprint(f' ├─────── ⏺ running', 'green')
            self.cprint(f' └─────── Container {self.container.short_id} {self.container.name}', 'green')
            self.cprint(f' ℹ Run the following commands in the terminal if you are debugging locally:', 'yellow')
            self.cprint(f' port=\"{self.port}\"', 'cyan')
            self.cprint(f' export JUPYTER_BASE_URL=\"{self.jh_url}\"', 'cyan')
            self.cprint(f' export JUPYTER_TOKEN=\"{self.jh_token}\"', 'cyan')
            self.cprint(f'␤[Resources]', 'light_grey')
            self.cprint(f' ├── MyST repository', 'magenta')
            self.cprint(f' │   ├───────── ✸ {self.rees.gh_user_repo_name}','light_blue')
            self.cprint(f' │   ├───────── ⎌ {self.rees.gh_repo_commit_hash}','light_blue')
            self.cprint(f" │   └───────── ⏲ {self.rees.repo_commit_info['datetime']}: {self.rees.repo_commit_info['message']}".replace('\n', ''),'light_blue')
            self.cprint(f' └── Docker container', 'magenta')
            self.cprint(f'     ├───────── ✸ {self.rees.pull_image_name}','light_blue')
            self.cprint(f'     ├───────── ⎌ {self.rees.binder_image_tag}','light_blue')
            self.cprint(f"     ├───────── ⏲ {self.rees.binder_commit_info['datetime']}: {self.rees.binder_commit_info['message']}".replace('\n', ''),'light_blue')
            if self.rees.binder_image_name:
                self.cprint(f'     └───────── ℹ Using NeuroLibre base image {self.rees.binder_image_name}','yellow')
            else:    
                self.cprint(f'     └───────── ℹ This image was built from REES-compliant {self.rees.gh_user_repo_name} repository at the commit above','yellow')
        except Exception as e:
            logging.error(f'Could not spawn a JH: \n {e}')

    def delete_stopped_containers(self):
        """
        Delete all stopped Docker containers.
        """
        stopped_containers = self.rees.docker_client.containers.list(all=True, filters={"status": "exited"})
        for container in stopped_containers:
            logging.info(f"Deleting stopped container: {container.id}")
            container.remove()

    def delete_image(self):
        """
        Delete the pulled Docker image.
        """
        if self.docker_image:
            logging.info(f"Deleting image: {self.docker_image.id}")
            self.rees.docker_client.images.remove(image=self.docker_image.id)

    def stop_container(self):
        """
        Stop and remove the running container.
        """
        if self.container:
            self.container.stop()
            self.container.remove()


from myst_libre.tools import JupyterHubLocalSpawner, MystMD
from myst_libre.rees import REES
from myst_libre.builders import MystBuilder



resources = REES(dict(
            registry_url="https://binder-registry.conp.cloud",
            gh_user_repo_name = "agahkarakuzu/mriscope",
            gh_repo_commit_hash = "ae64d9ed17e6ce66ecf94d585d7b68a19a435d70",
            binder_image_tag = "489ae0eb0d08fe30e45bc31201524a6570b9b7dd"))


hub = JupyterHubLocalSpawner(resources,
                             host_data_parent_dir = "/Users/agah/Desktop/tmp/DATA",
                             host_build_source_parent_dir = '/Users/agah/Desktop/tmp',
                             container_data_mount_dir = '/home/jovyan/data',
                             container_build_source_mount_dir = '/home/jovyan')

hub.spawn_jupyter_hub()

MystBuilder(hub).build_site()
